//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <scriptpp/scrvar.hpp>
#include <scriptpp/scrsubs.hpp>

#include "config.hpp"
#include "flog.hpp"
#include "state.hpp"
#include "runfile.hpp"
#include "textsubs.hpp"

#include "onewire.hpp"

// this MUST be kept in sync with the temp_sensor_config::wm_state enum
static const char * const watermark_state_names[] =
    { "undef", "low", "middle", "high", 0 };

struct temp_sensor : temp_sensor_config {
    int last_temp;
    unsigned long timemark;

    temp_sensor *next;

    temp_sensor()
        : last_temp(-1000), timemark(0), next(0) {}
    void TempChange(int temp);
    void InitWatermarks();  // to settle the configs earlier than the
                            // devices and the state is ready
    void StartWatermarks(); // to handle initial watermark events
    void ReportConfig();
private:
    void ChooseWatermarkEvent(const temp_sensor::watermark *p, int newtemp);
};

static temp_sensor *first_sensor = 0;

void init_onewire_info()
{
    int n = config_get_tsens_count();
    log_message("Total %d temp. sensor entries found", n);
    for(int i=0; i<n; i++) {
        temp_sensor *tp = new temp_sensor;
        if(config_get_tsens_config(i, *tp)) {
            tp->next = first_sensor;
            first_sensor = tp;
            tp->ReportConfig();
            tp->InitWatermarks();
        } else {
            delete tp;
        }
    }
}

void init_onewire_state()
{
    for(temp_sensor *t = first_sensor; t; t = t->next)
        t->StartWatermarks();
}

enum { ow_temp_crc_error = 0x7fff };

static bool owid_eq(unsigned char a[8], unsigned char b[8])
{
    int i;
    for(i=0; i<8; i++)
        if(a[i] != b[i]) return false;
    return true;
}

void handle_onewire_info(int /*id*/, unsigned char ow_id[8], int temperature)
{
    if(temperature == ow_no_temperature)
        return;   // current version only handles temp. sensors
    if(temperature == ow_temp_crc_error)
        return;   // simply ignore that

    for(temp_sensor *t = first_sensor; t; t = t->next) {
        if(owid_eq(ow_id, t->id)) {
            t->TempChange(temperature >> 4);
            return;   // done
        }
    }
}

static void make_watermark_dictionary(
    ScriptVector &dict,
    const temp_sensor *sensor,
    const temp_sensor::watermark *wmark,
    int new_temp
)
{
    dict.Clear();
    dict.AddItem("name");
    dict.AddItem(sensor->name);
    dict.AddItem("id");
    dict.AddItem(ScriptNumber(sensor->id_numerical));
    dict.AddItem("last_t");
    dict.AddItem(ScriptNumber(sensor->last_temp));
    dict.AddItem("t");
    dict.AddItem(ScriptNumber(new_temp));
}

static void
perform_watermark_event(const temp_sensor *sensor,
                        const temp_sensor::watermark *wmark,
                        int t, const char *comment,
                        const char *action)
{
    ScriptVector v;
    ScriptVector argv; /* remains empty here... */
    make_watermark_dictionary(v, sensor, wmark, t);
    ScriptVariable act_res = FFSubstitution(action, argv, v);
    if(t == -1000) // damn hack
        comment = "unknown";
    log_message("Temp. watermark %d/%d: t=%d (%s), doing [%s]",
        wmark->low_mark, wmark->high_mark, t, comment, act_res.c_str());
    state_process_event(act_res.c_str());
}

void temp_sensor::ChooseWatermarkEvent(const watermark *p, int new_temp)
{
    if(p->state == wms_undef) {
        switch(p->next_state) {
        case wms_high:
            if(p->high_init_event != "")
                perform_watermark_event(this, p, new_temp, "high",
                    p->high_init_event.c_str());
            break;
        case wms_low:
            if(p->low_init_event != "")
                perform_watermark_event(this, p, new_temp, "low",
                    p->low_init_event.c_str());
        case wms_undef:
        case wms_middle:
            ;
        }
    } else {
        switch(p->next_state) {
        case wms_high:
            perform_watermark_event(this, p, new_temp, "high",
                                    p->high_event.c_str());
            break;
        case wms_low:
            perform_watermark_event(this, p, new_temp, "low",
                                    p->low_event.c_str());
        case wms_undef:
        case wms_middle:
            ;
        }
    }
}

void temp_sensor::TempChange(int new_temp)
{
    int delta = new_temp - last_temp;
    if(new_temp > 80 || new_temp < -60 || 
        (last_temp != -1000 && (delta > 10 || delta < -10))
    ) {
        // just an error, to be ignored
        log_message("XXX (%s) last_temp = %d, new_temp = %d NONSENSE IGNORED",
                    name.c_str(), last_temp, new_temp);
        return;
    }
#if 0   // don't do this as this ruins the transit_counter mechanics
    if(delta == 0 || new_temp == saved_temp)
        return;
#endif
    if(delta != 0)
        log_message("XXX (%s) last_temp = %d, new_temp = %d",
                    name.c_str(), last_temp, new_temp);
    last_temp = new_temp; // it's okay, we don't use last_temp in the
                          // watermark handling, but it is what appears
                          // in the reports so let the new temperature
                          // appear there
    for(watermark *p = first; p; p = p->next) {
        wm_state new_temp_pos =
            (new_temp <= p->low_mark) ? wms_low :
            (new_temp >= p->high_mark) ? wms_high :
            wms_middle;
        if(p->next_state != wms_undef &&
            p->next_state != new_temp_pos &&
            p->transit_counter > 0
        ) {
            log_message("XXX (%s %d:%d) temperature ARTIFACT (%d) detected!",
                        name.c_str(), p->low_mark, p->high_mark,
                        p->transit_counter);
        }
        if(new_temp_pos == p->state) {
            // no change is to take place for this watermark
            p->next_state = wms_undef;
            continue;
        }
        if(new_temp_pos == wms_middle) {
            // we only become "middle" if we was in wms_undef
            // otherwise, we never change to the "middle" state
            if(p->state != wms_undef)
                continue;
        }
        // okay, change is to be processed
        if(p->next_state == new_temp_pos) {
            // repeated -> increment the counter
            p->transit_counter++;
        } else {
            // first time -> just start the counter
            p->next_state = new_temp_pos;
            p->transit_counter = 1;
        }
        log_message("XXX (%s) watermark %d/%d transit (%s->%s)[cnt=%d(%d)]",
            name.c_str(), p->low_mark, p->high_mark,
            watermark_state_names[p->state],
            watermark_state_names[new_temp_pos],
            p->transit_counter, p->transit_delay);
        // now check whether it is enough now
        if(p->transit_counter > p->transit_delay) {
            // YES! perform the change
            ChooseWatermarkEvent(p, new_temp);
            p->state = p->next_state;
            p->next_state = wms_undef;
            runfile_set_tempwm(name.c_str(), p->index,
                               watermark_state_names[p->state]);
            log_message("XXX (%s) watermark [%d] is now %s",
                        name.c_str(), p->index,
                        watermark_state_names[p->state]);
        }
    }
}

void temp_sensor::InitWatermarks()
{
    for(watermark *p = first; p; p = p->next) {
        ScriptVariable s =
            runfile_get_tempwm(name.c_str(), p->index);
        s.Tolower();
        p->state = wms_undef;
        for(int i = 0; watermark_state_names[i]; i++) {
            if(s == watermark_state_names[i]) {
                p->next_state = (wm_state)i;
                log_message("watermark %s:%d:%d will start in %s state",
                            name.c_str(), p->low_mark, p->high_mark,
                            s.c_str());
                break;
            }
        }
    }
}

void temp_sensor::StartWatermarks()
{
    for(watermark *p = first; p; p = p->next) {
        ChooseWatermarkEvent(p, -1000);
        p->state = p->next_state;
        p->next_state = wms_undef;
    }
}

void temp_sensor::ReportConfig()
{
    log_message("Temp. sensor: %s", name.c_str());
    for(watermark *p = first; p; p = p->next) {
        log_message("  +watermark: %d:%d:%s:%s:%d:%s:%s",
                    p->low_mark, p->high_mark,
                    p->low_event.c_str(), p->high_event.c_str(),
                    p->transit_delay,
                    p->low_init_event.c_str(), p->high_init_event.c_str());
    }
}

ScriptVariable onewire_report()
{
    ScriptVariable res("");
    for(temp_sensor *t = first_sensor; t; t = t->next) {
        if(t->last_temp != -1000) {
            if(res.Length() > 0)
                res += " ";
            res += ScriptVariable(0, "%s:%d", t->name.c_str(), t->last_temp);
        }
    }
    return res;
} 

int onewire_get_temperature(const char *sens_id)
{
    for(temp_sensor *t = first_sensor; t; t = t->next) {
        if(t->name == sens_id)
            return t->last_temp;
    }
    log_message("Temp. sensor %s not found, please check config", sens_id);
    return -1000;
}
