// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#include <string.h>
#include "sue_822.hpp"


SUEMessage::SUEMessage()
    : heading_line(0), header(), body()
{
}

SUEMessage::~SUEMessage()
{
    DropAll();
}

void SUEMessage::DropAll()
{
    if(heading_line)
        delete[] heading_line;
    heading_line = 0;
    header.DropAll();
    body.DropAll();
}

void SUEMessage::SetHeadingLine(const char *s)
{
    if(heading_line)
        delete[] heading_line;
    if(s) {
        heading_line = new char[strlen(s)+1];
        strcpy(heading_line, s);
    } else {
        heading_line = 0;
    }
}


void SUEMessage::Export(SUEBuffer &buf) const
{
    buf.DropAll();
    if(heading_line) {
        buf.AddString(heading_line);
        if(buf[buf.Length()-1] != '\r')
            buf.AddChar('\r');
        buf.AddChar('\n');
    }
    SUEStringMap::Iterator t(&header);
    for(; t.Name(); t.Next()) {
        buf.AddString(t.Name());
        buf.AddString(": ");
        buf.AddString(t.Body());
        if(buf[buf.Length()-1] != '\r')
            buf.AddChar('\r');
        buf.AddChar('\n');
    }
    buf.AddString("\r\n");
    buf.AddData(body.GetBuffer(), body.Length());
}


///////////////////////////////////////////////////

void SUEMessageIncoming::DropAll()
{
    SUEMessage::DropAll();
    state = start;
    last_error = err_ok;
    buf.DropAll();
}

bool SUEMessageIncoming::FeedChar(int c)
{
    bool res;
    switch(state) {
        case start:   res = HandleStart(c);   break;
        case heading: res = HandleHeading(c); break;
        case hname:   res = HandleHname(c);   break;
        case hbody:   res = HandleHbody(c);   break;
        case hskip0:  res = HandleHskip0(c);  break;
        case hskip:   res = HandleHskip(c);   break;
        case msgbody: res = HandleMsgbody(c); break;
        case finish:  res = HandleFinish(c);  break;
        case error:   res = false;
        default:      res = false;
    }
    if(res)
        last_error = err_ok;
    return res;
}

bool SUEMessageIncoming::IsOk() const
{
    return state == finish || state == msgbody || 
           ((state == hname || state == hskip ) && buf.Length() == 0);
}

bool SUEMessageIncoming::HandleStart(int c)
{
    if(c == -1) {   // empty message; why no?
        state = finish;
        return true;
    }
    if(c == '\r' || c == '\n' || c == '\t' || c == ' ')
        return true;
    // non-space char; handle it as the heading line char
    state = heading;
    return HandleHeading(c);
}

bool SUEMessageIncoming::HandleHeading(int c)
{
    if(c == -1) {
        state = error;
        last_error = err_incomplete_heading;
        return false;
    }
    if(c == '\r')
        return true;
    header_size++;
    if(c == '\n') {
        buf.AddChar(0);
        SetHeadingLine(buf.GetBuffer());
        buf.DropAll();
        state = hname;
        return true;
    }
    buf.AddChar(c);
    return true;
}

bool SUEMessageIncoming::HandleHname(int c)
{
    if(c == -1) {
        if(buf.Length() == 0) {
            state = finish;
            return true;
        } else {
            last_error = err_unexpected_eot;
            state = error;
            return false;
        } 
    }
    if(c == '\r' || c == ' ' || c == '\t') {
        last_error = err_space_in_header_name;
        return false;
    }
    if(c == '\n') {
        if(buf.Length() > 0) {
            last_error = err_no_colon_in_header;
            return false;
        } else {
            state = msgbody;
            return true;
        }
    }
    header_size++;
    if(c == ':') {
        colonpos = buf.Length();
        buf.AddChar(c);
        state = hbody;
        return true;
    }
    buf.AddChar(c);
    return true;
}

bool SUEMessageIncoming::HandleHbody(int c)
{
    if(c == -1) {
        last_error = err_header_line_incomplete;
        state = error;
        return false;
    }
    if(c == '\r')
        return true;
    if(c == '\n') {
        state = hskip0;
        return true;
    }
    header_size++;
    buf.AddChar(c);
    return true;
}

bool SUEMessageIncoming::HandleHskip0(int c)
{
    if(c == -1) {  // message with empty body; good.
        state = finish;
        return true;
    }
    if(c == ' ' || c == '\t' || c == '\r') {
        state = hskip;
        return true;
    }
    if(c == '\n') { // this means empty line
        FinalizeHeader();
        state = msgbody;
        return true;
    }
    // non-space char was first; it's simply the next header
    FinalizeHeader();
    buf.DropAll();
    state = hname;
    return HandleHname(c);
}

bool SUEMessageIncoming::HandleHskip(int c)
{
    if(c == -1) {  // message with empty body, having some wrong WS... accept.
        state = finish;
        return true;
    }
    if(c == ' ' || c == '\t' || c == '\r')
        return true;
    if(c == '\n') { // actually, this means empty line (with some WS in it)
                    // we accept it but it IS possible we shouldn't
        FinalizeHeader();
        state = msgbody;
        header_size++;
        return true;
    }
    // non-space; go on with the header
    header_size += 2;
    buf.AddChar(' ');
    buf.AddChar(c);
    state = hbody;
    return true;
}

bool SUEMessageIncoming::HandleMsgbody(int c)
{
    if(c == -1)
        state = finish;
    else
        GetBody().AddChar(c);
    return true;
}

bool SUEMessageIncoming::HandleFinish(int c)
{
    if(c != -1 && c != ' ' && c != '\r' && c != '\n' && c != '\r')
        state = error;
    return false;
}

void SUEMessageIncoming::FinalizeHeader()
{
    buf[colonpos] = 0;
    int i = colonpos + 1;
    while(buf[i] == ' ' || buf[i] == '\t')
        i++;
    buf.AddChar(0);
    AddHeader(buf.GetBuffer(), buf.GetBuffer() + i);
}
