// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#ifndef SENTRY_SUE_822_HPP
#define SENTRY_SUE_822_HPP

#include "sue_smap.hpp"
#include "sue_buf.hpp"

//! This class implements a rfc822-like message, suitable for SMTP and HTTP
/*! The object is intended to be constructed manually.  In order to read
    it from a stream, char by char, use the SUEMessageIncoming class.
 */
class SUEMessage {
    char *heading_line;
    SUEStringMap header;
    SUEBuffer body;

public:
    SUEMessage();
    ~SUEMessage();

    void DropAll();

    const char *GetHeadingLine() const { return heading_line; }
    void SetHeadingLine(const char *s);

    void AddHeader(const char *name, const char *body)
        { header.Add(name, body); }
    void SetHeader(const char *name, const char *body)
        { header.Set(name, body); }
    bool RemoveHeader(const char *name, bool all = false)
        { return header.Remove(name, all); }
    const char *GetHeader(const char *name) const { return header.Get(name); }
    const char *GetLastHeaderName() const { return header.GetLastName(); }
    const char *GetLastHeaderBody() const { return header.GetLastBody(); }
    int GetHeaderCount() const { return header.GetCount(); }

    SUEBuffer& GetBody() { return body; }
    const SUEBuffer& GetBody() const { return body; }

        //! Export the message to a buffer
        /*! The buffer is NOT cleared, so in effect this appends the
            message to the given buffer.  This is useful in Sessions
            allowing you to send the message directly to your
            outputbuffer object.
         */
    void Export(SUEBuffer &buf) const;
};

class SUEMessageIncoming : public SUEMessage {
    enum states {
        start,     // skipping whitespace expecting the heading line
        heading,   // reading the heading line, if any
        hname,     // reading header name, looking for ':'
        hbody,     // reading header body until '\n'
        hskip0,    // the first char after header: WS for continued header
        hskip,     // skipping space before the header continues
        msgbody,   // reading the message body (after the empty line)
        finish,    // EOT signaled when everything was okay
        error      // something erroneous happened
    } state;
    SUEBuffer buf;
    int colonpos;
    int header_size;
public:
    enum errors {
        err_ok,
        err_unexpected_eot,
        err_incomplete_heading,
        err_space_in_header_name,
        err_no_colon_in_header,
        err_header_line_incomplete
    };
private:
    errors last_error;
public:
    SUEMessageIncoming()
        : state(start), buf(), header_size(0), last_error(err_ok) {} 
    void DropAll();

    bool FeedChar(int c);
    bool FeedEOT() { return FeedChar(-1); }
    bool IsOk() const;
    bool HeaderEnded() const { return state == msgbody || state == finish; }
    int HeaderSize() const { return header_size; }
    int BodySize() const { return GetBody().Length(); }
    int TotalSize() const { return HeaderSize() + BodySize(); }
    errors GetLastError() const { return last_error; }
private:
    bool HandleStart(int c);
    bool HandleHeading(int c);
    bool HandleHname(int c);
    bool HandleHbody(int c);
    bool HandleHskip0(int c);
    bool HandleHskip(int c);
    bool HandleMsgbody(int c);
    bool HandleFinish(int c);
    void FinalizeHeader();
};

#endif
