// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#include <stdio.h> // for sprintf
#include <stdlib.h> // for rand
#include <string.h>
#include <time.h>
#include "md5.h"
#include "sue_sel.hpp"
#include "sue_822.hpp"
#include "sue_http.hpp"
#include "sue_digs.hpp"


SUEWebSession::SUEWebSession(SUEWebSessionCollection *c)
    : session_id(0), the_collection(c)
{
    SetFromNow(c->GetUnauthTimeout());
    c->GetSelector()->RegisterTimeoutHandler(this);
}

SUEWebSession::~SUEWebSession()
{
    if(the_collection)
        the_collection->SessionDeletionNotification(this);
    if(session_id)
        delete[] session_id;
}

enum { magic_number_1 = 0x1e6a1e3e, magic_number_2 = 0x5f10a7e4 };

char* SUEWebSession::GenerateSessionID(int idx)
{
    char *buf = new char[64];
    sprintf(buf, "%08x%08x%08x%08x",
                 ((idx * 117 + 135) ^ magic_number_1) & 0x7fffffff, 
                 rand() & 0xffffffff, (int)time(0) & 0xffffffff,
                 ((idx * 191 + 777) ^ magic_number_2) & 0x7fffffff);
    return buf;
}

int SUEWebSession::GetIdxBySessionID(const char *id)
{
    if(strlen(id) != 32)
        return -1;
    char buf[9];
    memcpy(buf, id, 8);
    buf[8] = 0;
    char *err;
    long i1 = strtol(buf, &err, 16);
    if(err != buf+8)
        return -1;
    i1 ^= magic_number_1;
    i1 -= 135;
    i1 /= 117;
    long i2 = strtol(id+24, &err, 16);
    if(err != id+32)
        return -1;
    i2 ^= magic_number_2;
    i2 -= 777;
    i2 /= 191;
    if(i1 != i2)
        return -1;
    return i1;
}

const char* SUEWebSession::GetSessionID(int idx) const
{
    if(!session_id)
        session_id = MakeSessionID(idx);
    return session_id;
}

int SUEWebSession::GetIdx() const
{
    return session_id ? GetIdxBySessionID(session_id) : -1;
}

void SUEWebSession::ProcessWebRequest(SUEHttpServerSession *sess,
        SUEMessage *request, SUEMessage *response, const char *username)
{
    HandleWebRequest(sess, request, response, username);
    the_collection->GetSelector()->RemoveTimeoutHandler(this);
    SetFromNow(the_collection->GetUnauthTimeout());
    the_collection->GetSelector()->RegisterTimeoutHandler(this);
}



SUEWebSessionCollection::SUEWebSessionCollection(SUEEventSelector *sel)
    : the_selector(sel), sessions(0), sess_len(0), sess_border(0), first(0)
{}

SUEWebSession* SUEWebSessionCollection::MakeWebSession()
{
    // first, pick index
    int idx = -1;
    if(first) { // pick one of them
        idx = first->idx;
        FreeItem *tmp = first;
        first = first->next;
        delete tmp;
    } else {  // no "fragmented" space, check the end...
        if(sess_border >= sess_len) { // time to resize the array
            int i;
            if(sessions) {
                SUEWebSession **old = sessions;
                sessions = new SUEWebSession*[sess_len * 2];
                for(i = 0; i < sess_border; i++)
                    sessions[i] = old[i];
                delete[] old;
                sess_len *= 2;
            } else {
                sess_len = 8;
                sessions = new SUEWebSession*[sess_len];
            }
            for(i = sess_border; i < sess_len; i++)
                sessions[i] = 0;
        }
        idx = sess_border;
        sess_border++;
    }
    sessions[idx] = DoMakeWebSession(idx);
    sessions[idx]->GetSessionID(idx);
    return sessions[idx];
}

void SUEWebSessionCollection::SessionDeletionNotification(SUEWebSession *sess)
{
    int idx = sess->GetIdx();
    if(idx >= 0 && idx < sess_border && sessions[idx] == sess) {
        sessions[idx] = 0;
        if(idx == sess_border - 1)
            sess_border--;
        else {
            // have to store it in the list of 'fragmented' free space
            FreeItem *p = new FreeItem;
            p->idx = idx;
            p->next = first;
            first = p;
        }
    }
}

SUEWebSession* SUEWebSessionCollection::FindSession(const char *sess_id) const
{
    int idx = SUEWebSession::GetIdxBySessionID(sess_id);
    if(idx < 0 || idx >= sess_border || !sessions || !sessions[idx])
        return 0;
    if(0 != strcmp(sess_id, sessions[idx]->GetSessionID()))
        return 0;
    return sessions[idx];
}

SUEWebSessionCollection::~SUEWebSessionCollection()
{
    int i;
    for(i = 0; i < sess_border; i++)
        if(sessions[i]) {
            sessions[i]->ForgetMaster();
            delete sessions[i];
        }
    delete[] sessions;
    while(first) {
        FreeItem *tmp = first;
        first = first->next;
        delete tmp;
    }
}


SUEDigestAuthMachine::
SUEDigestAuthMachine(SUEEventSelector *sel, const char *ar, const char *uri)
    : SUEWebSessionCollection(sel)
{
    realm = new char[strlen(ar)+1];
    strcpy(realm, ar);
    domains = new char[strlen(uri)+1];
    strcpy(domains, uri);
}
 
SUEDigestAuthMachine::~SUEDigestAuthMachine()
{
    delete[] realm;
    delete[] domains;
}


enum { md5strsize = 33 };

static char hex_digit(unsigned int i)
{
    i &= 0xf;
    if(i < 10) return i + '0';
    return i + ('a' - 10);
}

static void md5str_by_str(char *md5s, const char *s)
{
    unsigned char digest[MD5_DIGEST_SIZE];
    
    MD5StringHash(digest, s);
    int i, j;
    for(i = 0, j = 0; i < MD5_DIGEST_SIZE; i++, j+=2) {
        md5s[j] = hex_digit(digest[i] >> 4);
        md5s[j+1] = hex_digit(digest[i]);
    }
    md5s[md5strsize-1] = 0;
}
 
void SUEDigestAuthMachine::HandleHttpRequest(SUEHttpServerSession *sess,
                               SUEMessage *request, SUEMessage *respbuf)
{
    const char *auth_field = request->GetHeader("Authorization");
    if(!auth_field || strncasecmp("Digest ", auth_field, 7) != 0) {
        RequestAuth(sess, respbuf, false);
        return;
    } 
    SUEStringMapAuthParam sm(auth_field+7);

    const char *realm = sm.Get("realm", true);
    const char *nonce = sm.Get("nonce", true);
    const char *uri = sm.Get("uri", true);
    const char *response = sm.Get("response", true);
    const char *user = sm.Get("username", true);
    const char *password = user ? PasswdForUser(user) : 0;
    if(!realm || !nonce || !uri || !response || !user || !password) {
        RequestAuth(sess, respbuf, false);
        return;
    } 
    if(0 != strcmp(realm, this->realm)) {
        RequestAuth(sess, respbuf, false);
        return;
    } 

    // compute A1 of user:realm:password
    char hashA1[md5strsize];
    char *buf = new char[strlen(user) + strlen(realm) + strlen(password) + 3];
    sprintf(buf, "%s:%s:%s", user, realm, password);
    md5str_by_str(hashA1, buf);
    delete[] buf;

    // compute A2 of METHOD:URI
    char hashA2[md5strsize];
    const char *method = sess->GetMethod();
    buf = new char[strlen(uri) + strlen(method) + 2];
    sprintf(buf, "%s:%s", method, uri);
    md5str_by_str(hashA2, buf);
    delete[] buf;

    // compute response of A1:nonce:A2
    char computed_response[md5strsize];
    buf = new char[2*md5strsize + strlen(nonce) + 3];
    sprintf(buf, "%s:%s:%s", hashA1, nonce, hashA2);
    md5str_by_str(computed_response, buf);
    delete[] buf;

    if(0 != strcasecmp(response, computed_response)) {
        RequestAuth(sess, respbuf, false);
        return;
    } 

    SUEWebSession *websess = FindSession(nonce);
    if(!websess) {
        RequestAuth(sess, respbuf, true);
        return;
    }

    websess->ProcessWebRequest(sess, request, respbuf, user);
}

void SUEDigestAuthMachine::
RequestAuth(SUEHttpServerSession *sess, SUEMessage *response, bool stale)
{
    static const char format[] = "Digest realm=\"%s\", nonce=\"%s\"%s";
    SUEWebSession *websess = MakeWebSession();
    const char *sess_id = websess->GetSessionID();
    char *buf =
        new char[sizeof(format)+strlen(realm)+strlen(sess_id)+(stale?14:0)];
    sprintf(buf, format, realm, sess_id, stale ? ", stale=TRUE" : "");
    sess->AddResponseHeader("WWW-Authenticate", buf);
    sess->AbortSession(401, "Authorization Required");
    delete[] buf;
}




SUEStringMapAuthParam::SUEStringMapAuthParam(const char *from)
{
    enum states
        { start, inname, skipbefore, skipafter, intoken, qtoken, skipcomma };
    states state = start;
    const char *mark = from;
    char *name = 0, *value = 0;
    for(const char *p = from; *p; p++) {
        switch(state) {
        case start:
            if(*p != ' ' && *p != '\t') {
                mark = p;
                state = inname;
            }
            break;
        case inname:
            if(*p == ' ' || *p == '\t' || *p == '=') {
                int len = p - mark;
                name = new char[len+1];
                memcpy(name, mark, len);
                name[len] = 0;
                state = (*p == '=') ? skipafter : skipbefore;
            }
            break;
        case skipbefore:
            if(*p == '=')
                state = skipafter;
            break;
        case skipafter:
            if(*p != ' ' && *p != '\t') {
                if(*p == '\"') {
                    mark = p+1;
                    state = qtoken;
                } else {
                    mark = p;
                    state = intoken;
                }
            }
            break;
        case intoken:
        case qtoken:
            if((state == intoken && (*p == ' ' || *p == '\t' || *p == ','))
                || (state == qtoken && *p == '\"'))
            {
                int len = p - mark;
                value = new char[len+1];
                memcpy(value, mark, len);
                value[len] = 0;
                Add(name, value, true);
                name = value = 0;
                state = (state==intoken && *p==',') ? start : skipcomma;
            }
            break;
        case skipcomma:
            if(*p == ',')
                state = start;
        }
    }
    if(name && state == intoken) {
        Add(name, mark, false);
    }
    if(name)
        delete[] name;
    if(value)
        delete[] value;
}
