// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#ifndef SENTRY_SUE_DIGS_HPP
#define SENTRY_SUE_DIGS_HPP

#include "sue_sel.hpp"
#include "sue_smap.hpp"

class SUEHttpServerSession;
class SUEMessage;

class SUEWebSessionCollection;

//! Abstract tracked session of multiple requests
class SUEWebSession : public SUETimeoutHandler {
    mutable char *session_id;
    SUEWebSessionCollection *the_collection;
public:
    SUEWebSession(SUEWebSessionCollection *c);
    virtual ~SUEWebSession();

        //! Called by the master just before the master deletes this object
    void ForgetMaster() { the_collection = 0; }

        //! Process another request
        /*! This function is called by the Server Session once it is
            known that the request is made by the same user and belongs
            to the same session.  It calls, in turn, the HandleWebRequest
            function which is to be implemented by the user.
         */
    void ProcessWebRequest(SUEHttpServerSession *sess,
                           SUEMessage *request,
                           SUEMessage *response,
                           const char *username);

        //! Create the ID for this session
        /*! This version simply calls the static GenerateSessionID(),
            but it can as well be overridden to use some other generation
            algorithm
         */
    virtual char* MakeSessionID(int idx) const
        { return GenerateSessionID(idx); }

        //! Get the SessionID. 
        /*! If there's still no ID, it is generated using MakeSessionID()
            \param idx is the index of this session in the collection,
            whatever it is.  Actually it is only needed when the ID is
            generated; the index is ignored when there's already an ID.
         */  
    const char* GetSessionID(int idx = -1) const;

        //! Generate the SessionID string
        /*! \returns zero-terminated string allocated with new[];
            the string is to be handled and delete[]'d by the caller
         */
    static char* GenerateSessionID(int idx);

        //! 'Dig up' the index out of the ID
    static int GetIdxBySessionID(const char *id);

        //! Return the index, or -1 if not available
    int GetIdx() const;

protected:
        //! Handle the request
        /*! This is the place to implement whatever logic you write
            your server for, create and send the HTTP reply.
         */
    virtual void HandleWebRequest(SUEHttpServerSession *sess,
                                  SUEMessage *request,
                                  SUEMessage *response,
                                  const char *username) = 0;

private:
    virtual void TimeoutHandle() { delete this; }
};

class SUEWebSessionCollection {
    class SUEEventSelector *the_selector;
    SUEWebSession **sessions;
    int sess_len, sess_border;
    struct FreeItem {
        int idx;
        FreeItem *next;
    };
    FreeItem *first;
public:
    SUEWebSessionCollection(SUEEventSelector *selector);
    virtual ~SUEWebSessionCollection();

    void SessionDeletionNotification(SUEWebSession *sess);

    SUEWebSession* MakeWebSession();
    virtual SUEWebSession* DoMakeWebSession(int idx) = 0;

    SUEWebSession* FindSession(const char *sess_id) const;

    SUEEventSelector *GetSelector() const { return the_selector; }
    int GetUnauthTimeout() const { return 15; }  // numbers are to change
    int GetActiveTimeout() const { return 600; } //   in the future...
};

class SUEDigestAuthMachine : public SUEWebSessionCollection {
    char *realm;
    char *domains;
public:
        //! The constructor
        /*! \param realm is Authentication Realm as defined by rfc2617,
                   simply speaking it is the string displayed to the user,
                   identifying WHAT (s)he is logging in
            \param domains is a space-separated list of URIs served by
                   this AuthMachine, typically just absolute "local" paths
         */
    SUEDigestAuthMachine(SUEEventSelector *sel,
                         const char *realm, const char *domains); 
    ~SUEDigestAuthMachine(); 

        // remains pure virtual, to be given by the user
    // virtual SUEWebSession* DoMakeWebSession(int idx) = 0;

        //! Create or find the Web Session and pass the request to it
    void HandleHttpRequest(SUEHttpServerSession *sess,
                           SUEMessage *request,
                           SUEMessage *response);

    virtual const char *PasswdForUser(const char *user) const = 0;

private:
    void RequestAuth(SUEHttpServerSession *s, SUEMessage *resp, bool stale);
};


class SUEStringMapAuthParam : public SUEStringMap {
public:
    SUEStringMapAuthParam(const char *from);
};


#endif
